import collections
import itertools
import six

def flatten(sequence_of_sequences):
    return itertools.chain.from_iterable(sequence_of_sequences)


def pairwise(sequence):
    sequence = iter(sequence)
    try:
        while True:
            yield next(sequence), next(sequence)
    except StopIteration:
        pass


def ordered_map(sequence):
    return collections.OrderedDict(sorted(pairwise(sequence)))


class annotate(object):
    """
    Tells merky to tokenize a given object.

    Wrap any arbitrary object with this; the wrapper delegates
    to the wrapped object, but provides a `__merky__` attribute of
    value `True`, indicating to transformers that the object should
    be "tokenized".

    The object wrapped is accessible via `__merky_annotated__`.
    """
    __slots__ = ('__merky_annotated__',)
    __merky__ = True

    def __init__(self, wrapped):
        self.__merky_annotated__ = wrapped

    def __getattr__(self, attr):
        return getattr(self.__merky_annotated__, attr)

    def __iter__(self):
        return iter(self.__merky_annotated__)


def annotate_values(dictlike):
    """
    Returns a dictionary based on `dictlike` with each value wrapped by `annotate`.

    The dictionary itself is not wrapped by `annotate`, only the values.
    """
    return dict((k, annotate(v)) for k, v in six.iteritems(dictlike))




import collections

from nose import tools
from . import words
import merky

BASIC_SEQ = (words.SHEKELS, words.EUROS, words.EPEES)
BASIC_SEQ_JSON = words.unify('[',
                                '"', words.SHEKELS, '",',
                                '"', words.EUROS, '",',
                                '"', words.EPEES, '"',
                             ']')
BASIC_SEQ_HASH = '8129cb64b06e72be46eec5037eafc1583586325a'

UNANNOTATED_NESTED_SEQ = (words.ANGSTROM, BASIC_SEQ, 213, -315)
UNANNOTATED_NESTED_SEQ_JSON = words.unify('[',
                                            '"', words.ANGSTROM, '",',
                                            BASIC_SEQ_JSON, ',',
                                            '213,',
                                            '-315',
                                          ']')
UNANNOTATED_NESTED_SEQ_HASH = '14fe00c4e749eb96b8c66370902ac90a4f0d52fe'

BASIC_DICT = {words.SHEKELS: words.EUROS, words.ANGSTROM: words.EPEES}
BASIC_DICT_ORDERED = collections.OrderedDict((
                        (words.ANGSTROM, words.EPEES),
                        (words.SHEKELS, words.EUROS),
                     ))
BASIC_DICT_JSON = words.unify('{',
                                '"', words.ANGSTROM, '":"', words.EPEES, '",',
                                '"', words.SHEKELS, '":"', words.EUROS, '"',
                              '}')
BASIC_DICT_HASH = 'b22023acbbc3af979ba46d986706d8da7614f722'

UNANNOTATED_NESTED_DICT = {"dict": dict(BASIC_DICT), "list": BASIC_SEQ}
UNANNOTATED_NESTED_DICT_ORDERED = collections.OrderedDict((
                                    ("dict", BASIC_DICT_ORDERED),
                                    ("list", list(BASIC_SEQ))
                                  ))
UNANNOTATED_NESTED_DICT_JSON = words.unify('{',
                                            '"dict":', BASIC_DICT_JSON, ',',
                                            '"list":', BASIC_SEQ_JSON,
                                           '}')
UNANNOTATED_NESTED_DICT_HASH = '616bcc08fa64f4e8ed9cd7395c8981cdff045baf'
UNANNOTATED_NESTED_DICT_TOKENIZED = collections.OrderedDict((
                                      ("dict", BASIC_DICT_HASH),
                                      ("list", BASIC_SEQ_HASH)
                                     ))




def test_simple_unannotated_list():
    # It always does the top level guy regardless of annotation.
    t = merky.ExcludeAnnotationTransformer()
    r = t.transform(BASIC_SEQ)
    tools.assert_equal((BASIC_SEQ_HASH, list(BASIC_SEQ)), next(r))
    tools.assert_raises(StopIteration, next, r)


def test_nested_unannotated_list():
    # Top level always gets done, but this time, the nested guy wil be too.
    t = merky.ExcludeAnnotationTransformer()
    r = t.transform(UNANNOTATED_NESTED_SEQ)
    # We get the inner guy first, who is tokenized due to annotation
    tools.assert_equal((BASIC_SEQ_HASH, list(BASIC_SEQ)), next(r))
    # The outer guy is different this time, as the token replaced the nested list.
    # SHA1 of json of [words.ANGSTROM, BASIC_SEQ_HASH, 213, -315]
    tools.assert_equal(("3e2f759672d412ed84b7359e2613c94cfee250f2", [words.ANGSTROM, BASIC_SEQ_HASH, 213, -315]),
                       next(r))
    tools.assert_raises(StopIteration, next, r)


def test_nested_annotated_list():
    # And again, it always does the top level guy.  But since no members are annotated, it
    # will explode out anything else.
    t = merky.ExcludeAnnotationTransformer()
    r = t.transform((words.ANGSTROM, merky.annotate(BASIC_SEQ), 213, -315))
    tools.assert_equal((UNANNOTATED_NESTED_SEQ_HASH, [words.ANGSTROM, list(BASIC_SEQ), 213, -315]), next(r))
    tools.assert_raises(StopIteration, next, r)


def test_simple_unannotated_dict():
    t = merky.ExcludeAnnotationTransformer()
    r = t.transform(BASIC_DICT)
    tools.assert_equal((BASIC_DICT_HASH, BASIC_DICT_ORDERED), next(r))
    tools.assert_raises(StopIteration, next, r)


def test_nested_unannotated_dict():
    t = merky.ExcludeAnnotationTransformer()
    r = t.transform(UNANNOTATED_NESTED_DICT)
    # We get the inner dict first, who is tokenized due to lack of annotation
    tools.assert_equal((BASIC_DICT_HASH, BASIC_DICT_ORDERED), next(r))
    # We get the inner list second, who is tokenized due to lack of annotation
    tools.assert_equal((BASIC_SEQ_HASH, list(BASIC_SEQ)), next(r))
    # The outer guy is different this time, as tokens replace the nested dict and list
    # SHA1 of json of [words.ANGSTROM, BASIC_SEQ_HASH, 213, -315]
    tools.assert_equal((UNANNOTATED_NESTED_DICT_HASH, UNANNOTATED_NESTED_DICT_TOKENIZED),
                       next(r))
    tools.assert_raises(StopIteration, next, r)


def test_mixed_nesting():
    t = merky.ExcludeAnnotationTransformer()
    x = {"a": merky.annotate([
                    {"a": merky.annotate({"a": "A", "b": "B"}), "b": "B"},
                    {"nothing": "special"},
                ]),
         "b": [
                merky.annotate({"c": "C", "d": "D"}),
                {"e": "E", "f": "F"},
              ],
        }
    r = t.transform(x)
    # x["a"] is annotated and thus won't show up until the final structure.
    # The first guy will be x["b"][1], which does get tokenized.
    # Second guy will be x["b"], which also gets tokenized.
    # Final guy is the full container x.

    # First up: x["b"][1]
    tools.assert_equal(('37c3a03f839031d9a6eb3281b792a0cb6e02e79d',
                        collections.OrderedDict((
                            ('e', 'E'),
                            ('f', 'F'),
                        ))),
                       next(r))

    # Penultimate: x["b"]
    tools.assert_equal(('a6af006f34fec3ea34cf23b6505fd8eed6ccb3d6',
                        [collections.OrderedDict((
                            ('c', 'C'),
                            ('d', 'D')
                         )),
                         '37c3a03f839031d9a6eb3281b792a0cb6e02e79d']
                        ),
                       next(r))

    # Final, top level
    tools.assert_equal(('a723583cedba5006719024f7600301ea76287a36',
        collections.OrderedDict((
            ('a', [
                collections.OrderedDict((
                    ('a', collections.OrderedDict((('a', 'A'), ('b', 'B')))),
                    ('b', 'B'))),
                collections.OrderedDict((('nothing', 'special'),)),
                ]),
            ('b', 'a6af006f34fec3ea34cf23b6505fd8eed6ccb3d6')))),
        next(r))

    tools.assert_raises(StopIteration, next, r)




# Since not all python 3 versions support the "u" prefix, we just use the
# normal literal (which is unicode)
EUROS = "\u20ac20"
ANGSTROM = "\xc5ngstr\xf6m"
EPEES = "\xe9p\xe9es"
SHEKELS = "\u20aa20"

def unify(*segments):
    return ''.join(segments)




import json

def _json_default(item):
    try:
        return item.__merky_annotated__
    except:
        raise TypeError(
                "object of type %s is not JSON-serializable." % str(type(item)))

def json_serializer(sort=True):
    """
    Returns a function that can serialize to JSON, with unicode enabled,
    nans disallowed, keys sorted, and whitespace-free separators.

    The function will handle merky-annotated objects according to the particulars
    of the underlying annotated structure.

    You can override the sorted keys via `sort=False`.
    """
    return json.JSONEncoder(ensure_ascii=False,
                            allow_nan=False,
                            sort_keys=sort,
                            separators=(",",":"),
                            default=_json_default).encode




import collections

from nose import tools
from .. import samples
from merky import transformer
from merky.cases import walker

def s(sample_name, suffix):
    return getattr(samples, "%s_%s" % (sample_name, suffix))


def get_sample(sample_name):
    return [s(sample_name, suffix) for suffix
            in ('NATURAL', 'ORDERED', 'TOKENS')]


def reader_func(collection):
    return lambda t: collection[t]


def get_walker_inputs(structure):
    d = collections.OrderedDict(
            transformer.Transformer().transform(structure))

    # The "head" is the last token added to the collection.
    return reader_func(d), list(d.keys())[-1]


def scenario(sample_name):
    def decorator(func):
        def wrapper():
            natural, ordered, tokens = get_sample(sample_name)
            reader, head = get_walker_inputs(natural)
            target = walker.Walker(reader, head)
            return func(natural, ordered, tokens, target)
        wrapper.__name__ = func.__name__
        return wrapper
    decorator.__name__ = "scenario_%s" % sample_name
    return decorator


def has_structure(n, exp):
    tools.assert_equal(exp, n.structure)


def has_token(n, exp):
    tools.assert_equal(exp, n.token)


def extract(o, keys):
    for key in keys:
        o = o[key]
    return o


def match(tok, wlk, *keys):
    tok = extract(tok, keys)
    wlk = extract(wlk, keys)
    has_structure(wlk, tok.member)
    has_token(wlk, tok.token)


def unnested(wlk, *keys):
    tools.assert_raises(KeyError, extract, wlk, keys)


def matcher(name):
    def builder(nat, ord_, tok, wlk):
        return lambda *keys: match(tok, wlk, *keys)
    return scenario(name)(builder)()


def unnested_checker(name):
    def builder(nat, ord_, tok, wlk):
        return lambda *keys: unnested(wlk, *keys)
    return scenario(name)(builder)()


def test_list():
    m = matcher('WALKER_LIST_CASE')
    u = unnested_checker('WALKER_LIST_CASE')
    yield m
    yield u, 0 # This guy is a string.
    yield u, 1 # This guy is a number.
    yield m, 2 # This guy is a list.
    yield m, 3 # This guy is a dict.

    # Explore the nested list.
    yield u, 2, 0 # A string
    yield u, 2, 1 # A number
    yield m, 2, 2 # A dict.
    yield m, 2, 3 # A list.

    # The second-level dict is just strings.
    yield u, 2, 2, 'a'

    # The second-level list is also just strings.
    yield u, 2, 3, 0
    yield u, 2, 3, 1
    yield u, 2, 3, 2

    # Explore the first nested dict.
    yield u, 3, 'string' # A string
    yield u, 3, 'number' # A number
    yield m, 3, 'list' # A list
    yield m, 3, 'dict' # A dict
    
    # The second-level list is again just strings
    yield u, 3, 'list', 0
    yield u, 3, 'list', 1

    # And the second-level dict is as well.
    yield u, 3, 'dict', 'a'


def test_dict():
    m = matcher('WALKER_DICT_CASE')
    u = unnested_checker('WALKER_DICT_CASE')
    yield m
    yield m, 'list' # A list.
    yield m, 'dict' # A dict.
    yield u, 'string' # A string.
    yield u, 'number' # A number.

    # Explore the first level list.
    yield u, 'list', 0 # A string
    yield u, 'list', 1 # A number
    yield m, 'list', 2 # A dict
    yield m, 'list', 3 # A list.

    # Explore the second-tier dict.
    yield u, 'list', 2, 'string'
    yield u, 'list', 2, 'number'
    yield m, 'list', 2, 'list'
    yield m, 'list', 2, 'dict'
    # And the substructures are just strings.
    yield u, 'list', 2, 'list', 0
    yield u, 'list', 2, 'list', 1
    yield u, 'list', 2, 'dict', 'a'

    # Now explore the second-tier list.
    yield u, 'list', 3, 0 # A string
    yield u, 'list', 3, 1 # A number
    yield m, 'list', 3, 2 # A dict of string
    yield u, 'list', 3, 2, 'a'
    yield m, 'list', 3, 3 # A list of strings
    yield u, 'list', 3, 3, 0
    yield u, 'list', 3, 3, 1

    # Explore the first level dict.
    yield u, 'dict', 'string'
    yield u, 'dict', 'number'
    yield m, 'dict', 'list' # A list of strings
    yield u, 'dict', 'list', 0
    yield u, 'dict', 'list', 1
    yield m, 'dict', 'dict' # A dict of strings
    yield u, 'dict', 'dict', 'a'





import codecs
import collections
import json
from .. import serialization

class Structure(object):
    """
    Basic interface for a serializ(ed/able) "structure".

    Without providing specifics for read/write operations, gives
    the basic interface that these things should have in common.

    The purpose of the structure is to combine:
    * The map of token/structure pairs one gets from transformation
    * A "head" token indicating which structure is the top-level
      structure.

    The combination of the `head` token and the `get` method would
    allow for load of specific cases like the
    `merky.cases.attrgraph.AttributeGraph` and similar (the `from_token`
    classmethod).
    """
    def populate(self, token_structure_pairs):
        raise NotImplementedError("populate() not implemented.")

    def get(self, key):
        raise NotImplementedError("get() not implemented.")

    def __getitem__(self, key):
        raise NotImplementedError("__getitem()__ not implemented.")



class TokenMapStructure(Structure):
    """
    Maintains an internal in-memory map of token/structure pairs.

    While the `head` can be explicitly set, the `head` defaults to the
    last token seen via `populate()`.
    """
    @staticmethod
    def default_tokenmap():
        return collections.OrderedDict()

    def populate(self, token_structure_pairs):
        """
        Ingests the `token_structure_pairs` into the internal token map.

        Sets `head` to the last token seen, regardless of the current state of `head`.
        """
        self.tokenmap.update(token_structure_pairs)
        self.head = next(iter(reversed(self.tokenmap))) if len(self.tokenmap) > 0 else None

    def get(self, key):
        """
        Returns the corresponding structure from the internal map given the token `key`.
        """
        return self.tokenmap.get(key)

    
    def __getitem__(self, key):
        """
        Returns the corresponding structure from the internal map given the token 'key'.
        """
        return self.tokenmap[key]


class InMemoryStructure(TokenMapStructure):
    """
    An in-memory "store" that has both the read and write interfaces.
    """
    def __init__(self, tokenmap=None, head=None):
        self.tokenmap = self.default_tokenmap() if tokenmap is None else tokenmap
        self.head = head

    def close(self):
        pass


class JSONStreamReadStructure(TokenMapStructure):
    """
    Reads merkified structure from a utf-8 JSON stream.
    """
    def __init__(self, stream):
        self.stream = stream
        self.tokenmap, self.head = self.deserialize_from_stream(stream)

    @classmethod
    def deserialize_from_stream(cls, stream):
        return json.load(stream, encoding='utf-8')


class JSONStreamWriteStructure(TokenMapStructure):
    """
    Writes merkified structure to a utf-8 JSON stream.

    The instance accumulates state internally and only serializes to stream
    at close().
    """
    serializer = serialization.json_serializer(sort=False)

    def __init__(self, stream):
        self.tokenmap = self.default_tokenmap()
        self.stream = stream

    def serialize_to_stream(self, stream):
        stream.write(self.serializer([self.tokenmap, self.head]))

    def close(self):
        self.serialize_to_stream(self.stream)
        self.stream.flush()


class JSONFileReadStructure(JSONStreamReadStructure):
    """
    Reads merkified structure from utf-8 JSON file.
    """
    def __init__(self, path):
        self.path = path
        self.tokenmap, self.head = self.deserialize_from_file(self.path)

    @classmethod
    def deserialize_from_file(cls, path):
        with codecs.open(path, encoding="utf-8", mode="rb") as f:
            return cls.deserialize_from_stream(f)


class JSONFileWriteStructure(JSONStreamWriteStructure):
    """
    Writes merkified structure to utf-8 JSON file.

    The instance accumulates state internally and only serializes to the file
    at close().
    """
    def __init__(self, path):
        self.tokenmap = self.default_tokenmap()
        self.path = path

    def serialize_to_file(self, path):
        """
        Writes the current state to the `path` specified.
        """
        with codecs.open(path, encoding="utf-8", mode="wb") as f:
            self.serialize_to_stream(f)

    def close(self):
        """
        Writes the state out to the file at `self.path`.
        """
        self.serialize_to_file(self.path)





import hashlib

def hexdigest(encodable):
    return hashlib.sha1(encodable.encode("utf-8")).hexdigest()


def hexdigester(serializer):
    def inner(structure):
        return hexdigest(serializer(structure))
    return inner




class Walker(object):
    """
    Explore merky tokenized structures.

    Given a tokenized structure in the form of a reader function (which
    returns the normalized structure as a function of token) and a token,
    the `Walker` exposes the merkified structure as a graph:
        * the `token` attribute is the token
        * the `structure` attribute is the normalized structure
        * the act of looking up by key/index (e.g. `node[key]`) will return
          another `Walker` for the tokenized structure that is referenced
          by the structure at the specified key/index.

    For instance, suppose you have some list of two dicts:

        [{"a": "A"},
         {"b": "B"}]

    Suppose further that it has this merkified structure when transformed:

        # imagine that T0, T1, and T2 are actually SHA1 hexdigest tokens.
        ((T0, {"a": "A"}),
            (T1, {"b": "B"}),
            (T2, [T0, T1]))

    If you provide a `read` function that can look up the corresponding normalized structure
    for a given token (such that `read(T0)` would return `{"a": "A"}`, for instance), then
    a walker would be created:

        w = Walker(read, T2)

    You can now walk this thing top-down (based on the shape of the original list of dicts),
    but see the tokens and partial structures along the way.

        print w.token, w.structure # "T2 [T0, T1]"
        print w[0].token, w[0].structure # "T0 {'a': 'A'}"
        print w[1].token, w[1].structure # "T1 {'b': 'B'}"

    When using the `[]` mechanism, you are expected to know that the value at `w.structure[key]`
    is a token, meaning that the `Walker` will get that token and return a new `Walker` instance
    bound to the structure to which that token refers (looked up via the `read` function in this
    case).
    """
    __slots__ = ('reader', 'token', 'structure')

    def __init__(self, reader, token):
        self.reader = reader
        self.token = token
        self.structure = reader(token)


    def __getitem__(self, key):
        return self.node(self.structure[key])

    
    def node(self, key):
        return type(self)(self.reader, key)





import os
import tempfile
import shutil

from nose import tools

import merky
from merky.cases import attrgraph
from merky.cases import tokendict
from merky.store import structure

def tempdir(func):
    def wrapper(*p, **kw):
        dir_ = tempfile.mkdtemp()
        try:
            return func(*(p + (dir_,)), **kw)
        finally:
            shutil.rmtree(dir_)
    wrapper.__name__ = func.__name__
    return wrapper


def dag(attrs, members):
    return attrgraph.AttributeGraph(attrs=attrs, members=members)

def getpath(reader, *keys):
    current = reader.get(reader.head)
    for key in keys:
        current = reader.get(current[key])
    return current

@tempdir
def test_versioned_graph_save_and_restore(workdir):
    path = os.path.join(workdir, 'my-graph.json')
    static = dag({"unchanging": "eternal"}, {})

    graph_s0 = dag({"name": "graph", "version": 0}, {
                    "static": static,
                    "changing": dag({"a": "A"}, {}),
                    "removed": dag({"bleep": "blorp"}, {}),
                })

    graph_s1 = dag({"name": "graph", "version": 1}, {
                    "static": static,
                    "changing": dag({"b": "B"}, {}),
                })
    
    version_map = tokendict.TokenDict({"v0": graph_s0,
                                       "v1": graph_s0,
                                       "v2": graph_s1,
                                       "v3": graph_s1})

    transformer = merky.AnnotationTransformer()
    writer = structure.JSONFileWriteStructure(path)
    writer.populate(transformer.transform(version_map))
    writer.close()

    reader = structure.JSONFileReadStructure(path)
    tokens = tokendict.TokenDict.from_token(reader.head, reader.get)

    tools.assert_equal(["v0", "v1", "v2", "v3"], list(tokens.keys()))

    # Same states get same tokens.
    tools.assert_equal(tokens.dict_["v0"], tokens.dict_["v1"])
    tools.assert_equal(tokens.dict_["v2"], tokens.dict_["v3"])
    tools.assert_equal(getpath(reader, "v0", 1, "static"),
                       getpath(reader, "v2", 1, "static"))

    # Different states, different tokens.
    tools.assert_not_equal(tokens.dict_["v1"], tokens.dict_["v2"])
    tools.assert_not_equal(getpath(reader, "v0", 1, "changing"),
                           getpath(reader, "v2", 1, "changing"))


    restored = tokendict.TokenDict.from_token(
        reader.head,
        reader.get,
        attrgraph.AttributeGraph.from_token
    )

    tools.assert_equal(["v0", "v1", "v2", "v3"],
                        list(restored.keys()))

    tools.assert_equal(dict(graph_s0.attrs), dict(restored["v0"].attrs))
    tools.assert_equal(dict(graph_s1.attrs), dict(restored["v2"].attrs))
    tools.assert_equal(["changing", "removed", "static"],
                        list(sorted(restored["v0"].members.keys())))
    tools.assert_equal(["changing", "static"],
                        list(sorted(restored["v2"].members.keys())))
    tools.assert_equal(dict(graph_s0.members["changing"].attrs),
                        dict(restored["v1"].members["changing"].attrs))
    tools.assert_equal(dict(graph_s1.members["changing"].attrs),
                        dict(restored["v3"].members["changing"].attrs))
    tools.assert_equal(dict(restored["v0"].members["static"].attrs),
                        dict(restored["v3"].members["static"].attrs))




import codecs
import collections
import os
import tempfile
import shutil
from nose import tools
from six.moves import StringIO
from .. import words
from merky.store import structure

def od(*p, **kw):
    return collections.OrderedDict(*p, **kw)

class TestInMemoryStructure(object):

    HEAD_TOKEN = 'sometoptoken'
    LIST_TOKEN = 'somelisttoken'
    DICT_TOKEN = 'somedicttoken'
    HEAD = od((('a', 'somelisttoken'), ('b', 'somedicttoken')))
    DICT = od(((words.EUROS, words.SHEKELS), (words.EPEES, words.ANGSTROM)))
    LIST = ["some", "list", words.SHEKELS]

    PAIRS = (
            (LIST_TOKEN, LIST),
            (DICT_TOKEN, DICT),
            (HEAD_TOKEN, HEAD),
        )


    def get_read_store(self):
        return structure.InMemoryStructure(od(self.PAIRS), self.HEAD_TOKEN)

    
    def get_write_store(self):
        return structure.InMemoryStructure()


    def verify_dict(self, received, expect):
        tools.assert_equal(expect, received)


    def verify_list(self, received, expect):
        tools.assert_equal(expect, received)


    def verify_head(self, store, head):
        tools.assert_equal(head, store.head)


    def verify_write(self, store, head):
        # In-memory guy has same state before and after close() call.
        tools.assert_equal(od(self.PAIRS), store.tokenmap)
        store.close()
        tools.assert_equal(head, store.head)
        tools.assert_equal(od(self.PAIRS), store.tokenmap)


    def test_head_from_last_by_default(self):
        store = self.get_write_store()
        store.populate(iter(self.PAIRS))
        self.verify_head(store, self.HEAD_TOKEN)
        self.verify_write(store, self.HEAD_TOKEN)


    def test_overriddable_head(self):
        store = self.get_write_store()
        store.populate(iter(self.PAIRS))
        store.head = self.DICT_TOKEN
        self.verify_head(store, self.DICT_TOKEN)
        self.verify_write(store, self.DICT_TOKEN)


    def test_read_operations(self):
        store = self.get_read_store()
        self.verify_head(store, self.HEAD_TOKEN)
        self.verify_dict(store.get(self.HEAD_TOKEN), self.HEAD)
        self.verify_dict(store[self.HEAD_TOKEN], self.HEAD)
        self.verify_list(store.get(self.LIST_TOKEN), self.LIST)
        self.verify_list(store[self.LIST_TOKEN], self.LIST)
        self.verify_dict(store.get(self.DICT_TOKEN), self.DICT)
        self.verify_dict(store[self.DICT_TOKEN], self.DICT)
        tools.assert_equal(None, store.get('no no no'))
        tools.assert_raises(KeyError, lambda: store['no no no'])


q = lambda word: words.unify('"', word, '"')
def commify(*w):
    terms = []
    for word in w[:-1]:
        terms.append(word)
        terms.append(',')
    if len(w) > 0:
        terms.append(w[-1])
    return words.unify(*terms)

jl = lambda *w: words.unify('[', commify(*w), ']')
def jd(*pairs):
    return words.unify('{',
                            commify(*(words.unify(k, ':', w)
                                      for k, w in pairs)),
                        '}')


class TestStreamStructure(TestInMemoryStructure):
    def json(self, head):
        return jl(
            jd(
                (q(self.LIST_TOKEN), jl(*(q(w) for w in self.LIST))),
                (q(self.DICT_TOKEN), jd(*((q(k), q(v)) for k,v in self.DICT.items()))),
                (q(self.HEAD_TOKEN), jd(*((q(k), q(v)) for k,v in self.HEAD.items()))),
            ),
            q(head),
        )
    
    def get_read_store(self):
        self.stream = StringIO(self.json(self.HEAD_TOKEN))
        return structure.JSONStreamReadStructure(self.stream)

    def get_write_store(self):
        self.stream = StringIO()
        return structure.JSONStreamWriteStructure(self.stream)

    def verify_dict(self, received, expect):
        tools.assert_equal(dict(expect), received)

    def verify_list(self, received, expect):
        tools.assert_equal(list(expect), received)

    def verify_write(self, store, head):
        # Nothing written to stream until close().
        tools.assert_equal(0, self.stream.tell())
        tools.assert_equal('', self.stream.read())
        store.close()
        self.stream.seek(0)
        tools.assert_equal(self.json(head), self.stream.read())

    def test_no_close(self):
        store = self.get_read_store()
        tools.assert_raises(AttributeError, getattr, store, 'close')


class TestFileStructure(TestStreamStructure):
    def setup(self):
        self.workdir = tempfile.mkdtemp()
        self.path = os.path.join(self.workdir, "some-file.json")
    
    def teardown(self):
        shutil.rmtree(self.workdir)

    def get_read_store(self):
        with codecs.open(self.path, mode="wb", encoding="utf-8") as f:
            f.write(self.json(self.HEAD_TOKEN))
        return structure.JSONFileReadStructure(self.path)
            
    def get_write_store(self):
        return structure.JSONFileWriteStructure(self.path)

    def verify_write(self, store, head):
        # File doesn't even exist before close.
        tools.assert_false(os.path.exists(self.path))
        # But it will after close()
        store.close()
        tools.assert_true(os.path.exists(self.path))
        with codecs.open(self.path, mode='rb', encoding='utf-8') as f:
            tools.assert_equal(self.json(head), f.read())




import itertools
import six
from merky import util

class AttributeGraph(object):
    """
    Represent a hierarchy of nodes with attributes.

    An `AttributeGraph` can be thought of as a pair of dictionaries:
    * A dictionary of arbitrary key/value pairs; the "attributes" of a node.
    * A dictionary of "members"; the nodes that "belong" to the node or are logically "under" the node.

    The keys of the members dictionary identify the member nodes within that dictionary but do not have
    any other requirements (in terms of uniqueness) outside the scope of that dictionary.  Such concerns
    are left to the user.

    The values of the members dictionary are expected to be instances of `AttributeGraph`.  It is not
    required that they be; they will be annotated in any case.  However, when restoring a graph from
    a hash-to-structure mapping, they will be restored as `AttributeGraph` instances, so they must be of
    the appropriate shape.

    Both dictionaries will be annotated for use with the `merky.AnnotationTransformer`.  However, the
    expectation is that the "attributes" dictionary does not require further annotation (even if it has
    nested structures), while the values of the "members" dictionary will be annotated and thus tokenized.

    Suppose we have some logical structure like:

               -----------
               |  PARENT |
               | - attrs |
               -----------
                 /    \ 
            (has a)  (has a)
              /          \ 
             \/          \/
        -----------   -----------
        | CHILD_A |   | CHILD_B |
        | - attrs |   | - attrs |
        -----------   -----------

    We might represent that thus:

        AttributeGraph(
            attrs=PARENT_ATTRS_DICT,
            members={
                "a": AttributeGraph(attrs=CHILD_A_ATTRS_DICT),
                "b": AttributeGraph(attrs=CHILD_B_ATTRS_DICT),
            }
        )

    The logical annotation transformer output would look something like:

                               PARENT_TOKEN
                                     |
                                    \|
                                   PARENT: [PARENT_ATTRS_TOKEN, CHILDREN_TOKEN]
                                            /                      /
                                           /                      /
                                          /                      /
                                         /                      /
                                       |/                      /
        PARENT_ATTRS: OrderedDict([...])                      /
                                                             /
                                                           |/
            CHILDREN: OrderedDict([('a', CHILD_A_TOKEN), ('b', CHILD_B_TOKEN)])
                                         /                      |
                                        /                       |
                                      |/                        |/
        CHILD_A: [CHILD_A_ATTRS_TOKEN]      CHILD_B: [CHILD_B_ATTRS_TOKEN]
                    /                                      |
                  |/                                       |/
        CHILD_A_ATTRS: OrderedDict([...])   CHILD_B_ATTRS: OrderedDict([...])


    The basic rules for annotation (and thus annotated transformation):
    * The `AttributeGraph` will be represented as a list.
    * The first member of the list will be the tokenized `attrs` dictionary.
    * The optional second member of the list will be the tokenized `members`
      dictionary, the values of which are also tokenized; it's optional because
      it's excluded from the list when the dictionary is empty.
    """

    __slots__ = ('attrs', 'members')
    __merky__ = True

    def __init__(self, attrs=None, members=None):
        if attrs is None:
            attrs = self.get_default_attrs()
        if members is None:
            members = self.get_default_members()

        self.attrs = attrs
        self.members = members

    def __iter__(self):
        yield util.annotate(self.attrs)
        if len(self.members):
            yield util.annotate(util.annotate_values(self.members))


    @classmethod
    def get_default_attrs(cls):
        return {}


    @classmethod
    def get_default_members(cls):
        return {}

    
    @classmethod
    def attrs_from_token_list(cls, t_list, reader):
        return reader(t_list[0])


    @classmethod
    def members_from_token_list(cls, t_list, reader):
        if t_list is None or len(t_list) < 2:
            return None
        items = reader(t_list[1])
        return type(items)((k, cls.from_token(t, reader))
                           for k, t in six.iteritems(items))


    @classmethod
    def from_token(cls, token, reader):
        """
        Produces an `AttributeGraph` from the given `token` and `reader`.

        Parameters:
            `reader`:  a callable that, when given a "token" (the sha1 hash)
                       of a structure, returns the corresponding structure
                       (presumably from storage or cache or whatever).
            `token`:   the "token" corresponding to the list structure for the
                       `AttributeGraph`.

        If the structure has a non-empty members dictionary structure, each
        member (the values of the dictionary) will be recursively expanded
        using this same method.

        Returns an `AttributeGraph` with attributes and members fully realized.
        """
        tokens = reader(token)
        return cls(attrs=cls.attrs_from_token_list(tokens, reader),
                   members=cls.members_from_token_list(tokens, reader))


    def walk(self, prior, key, predicate=lambda o, p, k: True):
        chain = list(itertools.chain(((key, self),), prior))
        if predicate(self, prior, key):
            yield iter(chain)

        if self.members:
            for k, v in sorted(six.iteritems(self.members)):
                for result in v.walk(chain, k, predicate=predicate):
                    yield result

    def nodes(self):
        """
        Iterate over all nodes (and their ancestry) in the graph.

        Walks the AttributeGraph and per node, yields an iterator of
        `(key, node)` pairs.  That sequence of pairs is in depth-first
        order, starting with the specific node and its corresponding key
        within its parent, followed by the parent node and that node's key
        within its own parent, and so on all the way to the root of the tree.

        If you're only interested in each node in isolation, then you need
        only use the first item in each such group.  However, this model lets
        you see each node's "ancestry" or path through the hierarchy, which
        is helpful for various uses.

        The root node's key is always `None` within this scheme.

        Given some AttributeGraph A with child members B and C (and corresponding
        keys "b" and "c", respectively), then:

            [list(s) for s in A.nodes()]

        would produce:

            [
                [(None, A)],
                [("b", B), (None, A)],
                [("c", C), (None, A)],
            ]
    
        """

        return self.walk((), None)

    def leaves(self):
        """
        Iterate over all leaf nodes (and their ancestry) in the graph

        Like `nodes`, but only the sequences for leaf nodes are yielded; each
        sequence has the full ancestry, but the intermediary nodes do not
        yield a sequence.

        Given the A, B, C scenario described for `nodes()`,

            [list(s) for s in A.leaves()]

        would produce:

            [
                [("b", B), (None, A)],
                [("c", C), (None, A)],
            ]

        """
        return self.walk((), None, predicate=lambda o, p, k: (o.members is None or len(o.members) < 1))





from nose import tools
from .. import samples
from merky.cases import tokendict
import merky

NAT, ORD, TOK = (samples.TOKENDICT_CASE_NATURAL,
                 samples.TOKENDICT_CASE_ORDERED,
                 samples.TOKENDICT_CASE_TOKENS)


def test_dict_():
    d = {"foo": "bar", "baz": "biz"}
    g = tokendict.TokenDict(d)
    tools.assert_equal(d, g.dict_)


def test_get_and_set():
    d = {"foo": ["bar"]}
    conv = lambda v: list(v) if getattr(v, '__merky__', False) else None
    g = tokendict.TokenDict(d)
    tools.assert_equal(["bar"], conv(g.get("foo")))
    tools.assert_raises(KeyError, g.get, "blah")
    tools.assert_equal("default", g.get("blah", "default"))
    tools.assert_equal(["bar"], conv(g["foo"]))
    g["blah"] = ["somethin"]
    tools.assert_equal(["somethin"], conv(g.get("blah")))


def test_key_sort():
    d = {"z": "Z", "m": "M", "g": "G", "q": "Q", "c": "C", "w": "W", "a": "A"}
    g = tokendict.TokenDict(d)
    tools.assert_equal(list(sorted(d.keys())), [k for k, _ in g.items()])
    tools.assert_equal(list(sorted(d.keys())), list(g.keys()))


def test_value_annotation():
    d = {"a": ["a"], "b": ["b"], "c": ["c"]}
    g = tokendict.TokenDict(d)
    conv = lambda val: list(val) if getattr(val, '__merky__', False) else None

    tools.assert_equal(list(sorted(d.items())),
                       [(k, conv(v)) for k, v in g.items()])

    tools.assert_equal([d[k] for k in sorted(d.keys())],
                       [conv(v) for v in g.values()])



def test_tokenization():
    g = tokendict.TokenDict(NAT)
    i = merky.AnnotationTransformer().transform(g)
    tools.assert_equal((TOK["a"].token, ORD["a"]), next(i))
    tools.assert_equal((TOK["b"].token, ORD["b"]), next(i))
    tools.assert_equal((TOK["c"].token, ORD["c"]), next(i))
    tools.assert_equal((TOK["d"].token, ORD["d"]), next(i))
    tools.assert_equal((TOK.token, TOK.member), next(i))
    tools.assert_raises(StopIteration, next, i)


CACHE = dict((t.token, t.member)
             for t in (TOK["a"], TOK["b"], TOK["c"], TOK["d"], TOK))

def test_restoration():
    g = tokendict.TokenDict.from_token(TOK.token, CACHE.get)
    tools.assert_equal(["a", "b", "c", "d"], list(g.keys()))
    tools.assert_equal(ORD, g.dict_)


def test_restoration_with_builder():
    def wrapper(token, reader):
        def wrapped():
            return ('called me!', reader(token))
        return wrapped

    g = tokendict.TokenDict.from_token(TOK.token, CACHE.get, wrapper)
    tools.assert_equal(["a", "b", "c", "d"], list(g.keys()))
    tools.assert_equal(dict((k, ('called me!', v)) for k, v in ORD.items()),
                       dict((k, v()) for k, v in g.dict_.items()))





from nose import tools

from .. import samples
from merky.cases import attrgraph
from merky import transformer
from merky import util

def dag(attrs=None, members=None):
    return attrgraph.AttributeGraph(attrs, members)

def restore(reader, token):
    return attrgraph.AttributeGraph.from_token(token, reader)

def transform(subject):
    return transformer.AnnotationTransformer().transform(subject)

t = transform

def a(t):
    return util.annotate(t)

def accum_assert(expected, iterable):
    i = iter(iterable)
    accum = []
    for expect in expected:
        try:
            received = next(i)
            tools.assert_equal(expect, received)
            accum.append(received)
        except StopIteration:
            raise StopIteration("Expected: %s", repr(expect))
    tools.assert_raises(StopIteration, next, i)
    return accum

aa = accum_assert

def test_attributes_only():
    g = dag(attrs=samples.D_A_NATURAL)
    aa([(samples.D_A_TOKENS.token, samples.D_A_ORDERED),
        (samples.L_D_A_TOKENS.token, samples.L_D_A_TOKENS.member)],
        t(g))


def test_annotation():
    # This verifies that the attribute graph itself is "annotated" by default.
    g = [dag(attrs=samples.D_B_NATURAL)]
    i = iter(t(g))
    tools.assert_equal((samples.D_B_TOKENS.token, samples.D_B_ORDERED), next(i))
    # We're ignoring the hash of the list, just verifying that it exploded things out.
    graph = next(i)
    tools.assert_equal([samples.D_B_TOKENS.token], graph[1])
    # Again, don't care about hash, only that it references the previous
    tools.assert_equal([graph[0]], next(i)[1])
    # That's all, folks!
    tools.assert_raises(StopIteration, next, i)


def test_nested_attributes():
    g = dag(attrs=samples.D_A_IN_B_NATURAL)
    aa([(samples.D_A_IN_B_TOKENS.token, samples.D_A_IN_B_ORDERED), # Doesn't recurse attrs
         (samples.L_D_A_IN_B_TOKENS.token, samples.L_D_A_IN_B_TOKENS.member)],
        t(g))

def with_kids_scenario():
    nat = samples.L_NEST_PARTIAL_NATURAL
    ord_ = samples.L_NEST_PARTIAL_ORDERED
    tok = samples.L_NEST_PARTIAL_TOKENS
    return nat, ord_, tok, [
        (tok[0].token, ord_[0]), # top-level attrs
        (tok[1]["bro"][0].token, ord_[1]["bro"][0]), # bro attrs
        (tok[1]["bro"].token, tok[1]["bro"].member), # bro attr-only token list
        (tok[1]["sis"][0].token, ord_[1]["sis"][0]), # sis attrs
        (tok[1]["sis"].token, tok[1]["sis"].member), # sis attr-only token list
        (tok[1].token, tok[1].member), # top member token dict
        (tok.token, tok.member), # top attrs/members token list
    ]

def test_with_kids():
    nat, ord_, tok, with_kids_tokens = with_kids_scenario()

    g = dag(attrs=nat[0],
            members={
                "bro": dag(attrs=nat[1]["bro"][0]),
                "sis": dag(attrs=nat[1]["sis"][0]),
                })

    r = aa(with_kids_tokens, t(g))

def test_restore_with_kids():
    n, o, tok, with_kids_tokens = with_kids_scenario()
    token_store = dict(with_kids_tokens)
    head = with_kids_tokens[-1][0]
    g = restore(token_store.get, head)
    tools.assert_is_instance(g, attrgraph.AttributeGraph)
    tools.assert_equal(o[0], g.attrs)
    tools.assert_equal(["bro", "sis"], list(g.members.keys()))
    tools.assert_is_instance(g.members["bro"], attrgraph.AttributeGraph)
    tools.assert_is_instance(g.members["sis"], attrgraph.AttributeGraph)
    tools.assert_equal(n[1]["bro"][0], g.members["bro"].attrs)
    tools.assert_equal(n[1]["sis"][0], g.members["sis"].attrs)
    tools.assert_equal({}, g.members["bro"].members)
    tools.assert_equal({}, g.members["sis"].members)


def test_with_grandkids():
    n = samples.L_MULTI_NEST_PARTIAL_NATURAL
    o = samples.L_MULTI_NEST_PARTIAL_ORDERED
    t = samples.L_MULTI_NEST_PARTIAL_TOKENS

    g = dag(attrs=n[0],
            members={
                "1-0": dag(attrs=n[1]["1-0"][0],
                           members={
                               "2-0": dag(attrs=n[1]["1-0"][1]["2-0"][0]),
                               "2-1": dag(attrs=n[1]["1-0"][1]["2-1"][0]),
                               }),
                "1-1": dag(attrs=n[1]["1-1"][0],
                           members={
                               "2-0": dag(attrs=n[1]["1-1"][1]["2-0"][0]),
                               "2-1": dag(attrs=n[1]["1-1"][1]["2-1"][0]),
                               }),
                })

    r = aa(
        [
            (t[0].token, o[0]), # top attrs
            (t[1]["1-0"][0].token, o[1]["1-0"][0]), # 1-0 attrs
            (t[1]["1-0"][1]["2-0"][0].token, o[1]["1-0"][1]["2-0"][0]), # 1-0/2-0 attrs
            (t[1]["1-0"][1]["2-0"].token, t[1]["1-0"][1]["2-0"].member), # 1-0/2-0 attr-only token list
            (t[1]["1-0"][1]["2-1"][0].token, o[1]["1-0"][1]["2-1"][0]), # 1-0/2-1 attrs
            (t[1]["1-0"][1]["2-1"].token, t[1]["1-0"][1]["2-1"].member), # 1-0/2-1 attr-only token list
            (t[1]["1-0"][1].token, t[1]["1-0"][1].member), # 1-0 member token dict
            (t[1]["1-0"].token, t[1]["1-0"].member), # 1-0 attr/member token list
            (t[1]["1-1"][0].token, o[1]["1-1"][0]), # 1-1 attrs
            (t[1]["1-1"][1]["2-0"][0].token, o[1]["1-1"][1]["2-0"][0]), # 1-1/2-0 attrs
            (t[1]["1-1"][1]["2-0"].token, t[1]["1-1"][1]["2-0"].member), # 1-1/2-0 attr-only token list
            (t[1]["1-1"][1]["2-1"][0].token, o[1]["1-1"][1]["2-1"][0]), # 1-1/2-1 attrs
            (t[1]["1-1"][1]["2-1"].token, t[1]["1-1"][1]["2-1"].member), # 1-1/2-1 attr-only token list
            (t[1]["1-1"][1].token, t[1]["1-1"][1].member), # 1-1 member token dict
            (t[1]["1-1"].token, t[1]["1-1"].member), # 1-1 attr/member token list
            (t[1].token, t[1].member), # top member token dict
            (t.token, t.member), # top attr/member token list
        ],
        transform(g))

def walker_graph():
    return dag(
        attrs={"name": "grandparent"},
        members={
            "a": dag(
                attrs={"name": "parent-a"},
                members={
                    "a": dag(attrs={"name": "child-a-a"}),
                    "b": dag(attrs={"name": "child-a-b"}),
                }
            ),
            "b": dag(
                attrs={"name": "parent-b"},
                members={
                    "a": dag(attrs={"name": "child-b-a"}),
                    "b": dag(attrs={"name": "child-b-b"}),
                }
            ),
        }
    )


def verify_names(expect, i):
    for keynameiter in expect:
        received = next(i)
        tools.assert_equal(list(keynameiter), [(k, m.attrs["name"]) for k, m in received])
    tools.assert_raises(StopIteration, next, i)

def test_node_walk():
    top = walker_graph()
    verify_names(
        [
            [(None, "grandparent")],
            [("a", "parent-a"), (None, "grandparent")],
            [("a", "child-a-a"), ("a", "parent-a"), (None, "grandparent")],
            [("b", "child-a-b"), ("a", "parent-a"), (None, "grandparent")],
            [("b", "parent-b"), (None, "grandparent")],
            [("a", "child-b-a"), ("b", "parent-b"), (None, "grandparent")],
            [("b", "child-b-b"), ("b", "parent-b"), (None, "grandparent")],
        ],
        top.nodes()
    )

def test_leaf_walk():
    top = walker_graph()
    verify_names(
        [
            [("a", "child-a-a"), ("a", "parent-a"), (None, "grandparent")],
            [("b", "child-a-b"), ("a", "parent-a"), (None, "grandparent")],
            [("a", "child-b-a"), ("b", "parent-b"), (None, "grandparent")],
            [("b", "child-b-b"), ("b", "parent-b"), (None, "grandparent")],
        ],
        top.leaves()
    )




import six
from merky import util

NO_DEFAULT=object()

class TokenDict(object):
    """
    A `dict`-like structure that annotates its values and ensures order.

    While the `TokenDict` looks and acts like a `dict`, its values are always
    annotated.  To access the "real" values, use the `dict_` member, which is
    the underlying dictionary itself.

    You can mutate that dictionary directly, and subsequent calls to
    `items`, `iteritems`, `keys`, `values` on the `TokenDict` will reflect
    those changes.

    As always, mutate with discretion.
    """
    __merky__ = True

    def __init__(self, dict_):
        self.dict_ = dict_

    def get(self, item, default=NO_DEFAULT):
        """
        Returns the annotated value associated with `key`.

        Raises a `KeyError` if the key is missing, unless a `default` is
        provided.  When the default is used, it is returned as-is without
        additional annotation.
        """
        try:
            return util.annotate(self.dict_[item])
        except KeyError:
            if default is NO_DEFAULT:
                raise
            return default


    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, val):
        self.dict_[key] = val

    def _pairs(self):
        return sorted(six.iteritems(self.dict_))

    def keys(self):
        """
        Iterates over the keys in sorted order.
        """
        return iter(k for k, _ in self._pairs())

    def values(self):
        """
        Iterates over the annotated values in sorted-key order.
        """
        return iter(util.annotate(v) for _, v in self._pairs())

    def items(self):
        """
        Iterates over the `(key, value)` pairs in sorted-key order.
        """
        for k, v in self._pairs():
            yield k, util.annotate(v)

    def iteritems(self):
        """
        See `items()`.
        """
        return self.items()

    @classmethod
    def from_token(cls, token, reader, builder=None):
        """
        Assembles a `TokenDict` by walking a graph of tokens.

        Parameters:
            `token`:   the "token" identifying the top-level structure
                       to be converted to a `TokenDict`.
            `reader`:  a function that, given a "token", returns the
                       corresponding structure.
            `builder`: (Optional) a function to apply to each token in the
                       `TokenDict` when assembling; it will be called with the
                       token seen and the `reader` above.  Ue this to, for instance,
                       convert each value into a more specific object.

        Returns a new `TokenDict` instance.
        """
        dict_ = reader(token)
        handler = reader if builder is None else lambda val: builder(val, reader)
        return cls(dict((k, handler(v)) for k, v in six.iteritems(dict_)))




import collections

from nose import tools
from . import words
import merky

BASIC_SEQ = (words.SHEKELS, words.EUROS, words.EPEES)
BASIC_SEQ_JSON = words.unify('[',
                                '"', words.SHEKELS, '",',
                                '"', words.EUROS, '",',
                                '"', words.EPEES, '"',
                             ']')
BASIC_SEQ_HASH = '8129cb64b06e72be46eec5037eafc1583586325a'

UNANNOTATED_NESTED_SEQ = (words.ANGSTROM, BASIC_SEQ, 213, -315)
UNANNOTATED_NESTED_SEQ_JSON = words.unify('[',
                                            '"', words.ANGSTROM, '",',
                                            BASIC_SEQ_JSON, ',',
                                            '213,',
                                            '-315',
                                          ']')
UNANNOTATED_NESTED_SEQ_HASH = '14fe00c4e749eb96b8c66370902ac90a4f0d52fe'

BASIC_DICT = {words.SHEKELS: words.EUROS, words.ANGSTROM: words.EPEES}
BASIC_DICT_ORDERED = collections.OrderedDict((
                        (words.ANGSTROM, words.EPEES),
                        (words.SHEKELS, words.EUROS),
                     ))
BASIC_DICT_JSON = words.unify('{',
                                '"', words.ANGSTROM, '":"', words.EPEES, '",',
                                '"', words.SHEKELS, '":"', words.EUROS, '"',
                              '}')
BASIC_DICT_HASH = 'b22023acbbc3af979ba46d986706d8da7614f722'

UNANNOTATED_NESTED_DICT = {"dict": dict(BASIC_DICT), "list": BASIC_SEQ}
UNANNOTATED_NESTED_DICT_ORDERED = collections.OrderedDict((
                                    ("dict", BASIC_DICT_ORDERED),
                                    ("list", list(BASIC_SEQ))
                                  ))
UNANNOTATED_NESTED_DICT_JSON = words.unify('{',
                                            '"dict":', BASIC_DICT_JSON, ',',
                                            '"list":', BASIC_SEQ_JSON,
                                           '}')
UNANNOTATED_NESTED_DICT_HASH = '7e5c9d9334702c025e185a9f63e380d16b85b083'


def test_simple_unannotated_list():
    # It always does the top level guy regardless of annotation.
    t = merky.AnnotationTransformer()
    r = t.transform(BASIC_SEQ)
    tools.assert_equal((BASIC_SEQ_HASH, list(BASIC_SEQ)), next(r))
    tools.assert_raises(StopIteration, next, r)


def test_nested_unannotated_list():
    # And again, it always does the top level guy.  But since no members are annotated, it
    # won't explode out anything else.
    t = merky.AnnotationTransformer()
    r = t.transform(UNANNOTATED_NESTED_SEQ)
    tools.assert_equal((UNANNOTATED_NESTED_SEQ_HASH, [words.ANGSTROM, list(BASIC_SEQ), 213, -315]), next(r))
    tools.assert_raises(StopIteration, next, r)

def test_nested_annotated_list():
    # Top level always gets done, but this time, the nested guy wil be too.
    t = merky.AnnotationTransformer()
    r = t.transform((words.ANGSTROM, merky.annotate(BASIC_SEQ), 213, -315))
    # We get the inner guy first, who is tokenized due to annotation
    tools.assert_equal((BASIC_SEQ_HASH, list(BASIC_SEQ)), next(r))
    # The outer guy is different this time, as the token replaced the nested list.
    # SHA1 of json of [words.ANGSTROM, BASIC_SEQ_HASH, 213, -315]
    tools.assert_equal(("3e2f759672d412ed84b7359e2613c94cfee250f2", [words.ANGSTROM, BASIC_SEQ_HASH, 213, -315]),
                       next(r))
    tools.assert_raises(StopIteration, next, r)

def test_simple_unannotated_dict():
    t = merky.AnnotationTransformer()
    r = t.transform(BASIC_DICT)
    tools.assert_equal((BASIC_DICT_HASH, BASIC_DICT_ORDERED), next(r))
    tools.assert_raises(StopIteration, next, r)


def test_nested_unannotated_dict():
    t = merky.AnnotationTransformer()
    r = t.transform(UNANNOTATED_NESTED_DICT)
    tools.assert_equal((UNANNOTATED_NESTED_DICT_HASH, UNANNOTATED_NESTED_DICT_ORDERED),
                       next(r))
    tools.assert_raises(StopIteration, next, r)


def test_mixed_nesting():
    t = merky.AnnotationTransformer()
    x = {"a": merky.annotate([
                    {"a": merky.annotate({"a": "A", "b": "B"}), "b": "B"},
                    {"nothing": "special"},
                ]),
         "b": [
                merky.annotate({"c": "C", "d": "D"}),
                {"e": "E", "f": "F"},
              ],
        }
    r = t.transform(x)
    # Lowest level: x["a"][0]["a"]
    tools.assert_equal(('5985f150e6c8051a45ba0082f0724ef983a56bc5',
                            collections.OrderedDict((('a', 'A'), ('b', 'B')))),
                       next(r))

    # Next up: x["a"]
    tools.assert_equal(('13b6b33de8c689ad12f99e2974a4f25ea4260aa9',
                            [
                                collections.OrderedDict((
                                    ('a', '5985f150e6c8051a45ba0082f0724ef983a56bc5'),
                                    ('b', 'B'),
                                )),
                                collections.OrderedDict((
                                    ('nothing', 'special'),
                                )),
                            ]),
                        next(r))

    # Penultimate: x["b"][0]
    tools.assert_equal(('4b78de319d1cf5639a8eb396a8cd6cdbb9fe784c',
                        collections.OrderedDict((
                            ('c', 'C'),
                            ('d', 'D'),
                        ))),
                       next(r))

    # Final, top level
    tools.assert_equal(('c6d5f854dca6a1e46d2d1c1d91aaab21bd484359',
                        collections.OrderedDict((
                            ('a', '13b6b33de8c689ad12f99e2974a4f25ea4260aa9'),
                            ('b', ['4b78de319d1cf5639a8eb396a8cd6cdbb9fe784c',
                                    collections.OrderedDict((
                                        ('e', 'E'),
                                        ('f', 'F'),
                                    )),
                                  ]),
                        ))),
                       next(r))

    tools.assert_raises(StopIteration, next, r)




import six

from . import util

def string_handler(item):
    if hasattr(item, "encode"):
        return item, None, False
    raise TypeError


def map_handler(item):
    try:
        i = util.flatten(sorted(six.iteritems(item)))
        return i, util.ordered_map, True
    except AttributeError:
        raise TypeError


def seq_handler(item):
    return iter(item), list, True


def default_handler(item):
    return item, None, False


def annotation_handler(handler, name=None, doc=None):
    def wrapped(item):
        i, collector, _ = handler(item)
        return i, collector, getattr(item, '__merky__', False)
    if name:
        wrapped.__name__ = name
    if doc:
        wrapped.__doc__ = doc
    return wrapped



def dispatcher(*handlers):
    def inner(item):
        for handler in handlers:
            try:
                return handler(item)
            except TypeError:
                pass
        raise TypeError("No handler could be found for item of type %s" % type(item))
    return inner


def walker(structure, dispatcher, tokenizer):
    stack = []
    accum = []
    top_flag = True
    current, collector, tokenize = iter((structure,)), lambda x: x, False
    while True:
        try:
            next_item, next_col, next_tok = dispatcher(next(current))
            if next_col:
                stack.append((current, collector, tokenize, accum))
                current, collector, tokenize, accum, top_flag = next_item, \
                                                                next_col, \
                                                                (next_tok or top_flag), \
                                                                [], \
                                                                False
            else:
                accum.append(next_item)
        except StopIteration:
            value = collector(accum)
            if tokenize:
                t = tokenizer(value)
                yield (t, value)
                value = t

            if not stack:
                break

            current, collector, tokenize, accum = stack.pop()
            accum.append(value)


map_annotation_handler = annotation_handler(map_handler, 'map_annotation_handler')
seq_annotation_handler = annotation_handler(seq_handler, 'seq_annotation_handler')

full_nesting_dispatcher = dispatcher(string_handler,
                                     map_handler,
                                     seq_handler,
                                     default_handler)

annotation_dispatcher = dispatcher(string_handler,
                                   map_annotation_handler,
                                   seq_annotation_handler,
                                   default_handler)

def excluder_handler(handler, converter, collector, name):
    def handle(item):
        next_item, next_col, next_tok = handler(item)
        if next_tok:
            next_item = converter(next_item)
            next_col = collector
        return next_item, next_col, not next_tok
    handle.__name__ = name
    return handle

def _unmerk(items):
    for i in items:
        yield getattr(i, '__merky_annotated__', i)

exclude_annotation_dispatcher = dispatcher(
        string_handler,
        excluder_handler(
            map_annotation_handler,
            lambda m: util.flatten((k, util.annotate(v))
                for k, v in util.pairwise(m)),
            lambda i: util.ordered_map(_unmerk(i)),
            'map_exclude_handler'),
        excluder_handler(
            seq_annotation_handler,
            lambda s: (util.annotate(v) for v in s),
            lambda i: list(_unmerk(i)),
            'seq_exclude_handler'),
        default_handler)




from setuptools import setup
import os

setup(
    name = "merky",
    version = "0.1.0",
    author = "Ethan Rowe",
    author_email = "ethan@the-rowes.com",
    description = ("JSON-oriented merkle tree utilities"),
    license = "MIT",
    url = "https://github.com/ethanrowe/python-merky",
    packages = ["merky",
                "merky.cases",
                "merky.store",
                "merky.test",
                "merky.test.misc",
                "merky.test.store",
                "merky.test.usecases",
    ],
    long_description = """
Merky - compute merkle trees for JSON-friendly data.
""",
    test_suite = "nose.collector",
    install_requires = [
                'six >= 1.5',
    ],
    setup_requires = [
        'nose',
        'mock >= 1.0.1',
    ],
    tests_require = [
        'nose',
        'mock >= 1.0.1',
    ],
)



from . import digest
from . import serialization
from . import tree

class Transformer(object):
    """
    Basic merkle tree inpired transform class.

    A `Transformer` can be given a nested data structure of arbitrary
    complexity, and will:
    - Walk the structure depth first
    - Calculate the sha1 (hex digest) of each distinct substructure
    - Yield the ``(sha1, substructure)`` pair to the caller
    - Substitute the sha1 for that substructure within its containing structure
    
    The top-level structure given is always yielded with its hash, whether it
    was complex or not.

    The depth-first walk will traverse anything that looks like a dict
    (that is, it implements ``iteritems``) or things that support
    iteration (implement ``__iter__``), with the exception of strings.  Strings
    are identified as anything that implements ``encode``.  Note that this means
    a ``json.JSONEncoder`` would be treated as a string, for example.

    Each structure is converted into a normal form, such that:
    * Strings, numbers, and booleans are left as is.
    * Dict-like structures get sorted by key and converted to
      ``collections.OrderedDict``.
    * Sequences get converted to `list` objects.

    This gives a result that can in principle produce a deterministic serialized
    form suitable for sha1 hashing.  Note that floating point values cannot be
    relied upon to give a deterministic result, and consequently they should be
    handled with care; consider converting them to string representations of
    fixed scale first.

    JSON is used for the serialization, and allows for unicode characters in
    strings and disallows the use of nan.
    """
    def __init__(self):
        self.serializer = self.get_serializer()
        self.tokenizer = self.get_tokenizer()
        self.dispatcher = self.get_dispatcher()

    def walker(self, structure):
        """
        Returns the `merky.tree.walker` for `structure` given self's `dispatcher` and `tokenizer`.
        """
        return tree.walker(structure, self.dispatcher, self.tokenizer)


    def transform(self, structure):
        """
        Walks the `structure` depth-first and yields `(sha1, normal_structure)` pairs according to
        the rules of `self.dispatcher`.

        If the same substructure appears multiple times within `structure`, it should be yielded
        with its corresponding sha1 hash the same number of times; duplicates are not filtered.
        """
        seen = False
        for pair in self.walker(structure):
            yield pair
            seen = True
        if not seen:
            yield (self.tokenizer(structure), structure)


    def get_serializer(self):
        """
        Returns a function to be used for structure serialization.  This implementation uses
        JSON with unicode support, no nans, and sorted keys.  The serialization must be consistent
        between uses to expect a consistent hash for the same input structure.

        Subclasses could override this to use an alternate serialization approach; the result must
        support `encode()`.
        """
        return serialization.json_serializer()


    def get_dispatcher(self):
        """
        Returns a function to use as the type-handler dispatcher within the object graph walk.
        This implementation uses the `merky.tree.full_nesting_dispatcher`.
        """
        return tree.full_nesting_dispatcher


    def get_tokenizer(self):
        """
        Returns a function to use to convert structures to "tokens" (hashes); this assumes that
        the serializer from `get_serializer` is already available at `self.serializer`.

        This implementation uses the JSON serializer of `get_serializer` and the sha1 hexdigest
        of `merky.digest.hexdigest`.
        """
        return digest.hexdigester(self.serializer)


class AnnotationTransformer(Transformer):
    """
    A merky Transformer that normalizes the input structure, but only yields/tokenizes
    structures that have been "annotated" with the ``__merky__`` attribute with a
    true value.

    The top-level structure is always tokenized and yielded regardless of the presence of the
    annotation.

    See the `merky.util.annotate` helper for annotating your data structures for use with
    this transformer.
    """

    def get_dispatcher(self):
        return tree.annotation_dispatcher


class ExcludeAnnotationTransformer(Transformer):
    """
    A merky Transformer that normalizes the input structure, but excludes
    structures that have been "annotated" with the ``__merky__`` attribute with a
    true value.

    The top-level structure is always tokenized and yielded regardless of the presence of the
    annotation.

    See the `merky.util.annotate` helper for annotating your data structures for use with
    this transformer.
    """
    def get_dispatcher(self):
        return tree.exclude_annotation_dispatcher



import collections
import hashlib
import json
import sys
import six

COMMON_CODE = """
from collections import OrderedDict

class Tokenset(object):
    def __init__(self, tokens, token):
        self.tokens = tokens
        self.token = token
        self.member = tokens[token]

    def __getitem__(self, key):
        return Tokenset(self.tokens, token=self.member[key])

    def __repr__(self):
        return "Tokenset(%s, token=%s)" % (repr(self.tokens), repr(self.token))

    def __str__(self):
        return self.__repr__()

"""

exec(COMMON_CODE)

def op_or_ident(target, op, **passthru):
    try:
        return getattr(target, op)(**passthru)
    except AttributeError:
        return target

JSON_ENCODER = json.JSONEncoder(allow_nan=False,
                                sort_keys=True,
                                separators=(",",":"),
                                ensure_ascii=False)


def tokenize(value):
    j = JSON_ENCODER.encode(value)
    return hashlib.sha1(j.encode('utf-8')).hexdigest()


class Node(object):
    def __init__(self, vals, recurse=True):
        self.vals = vals
        self.recurse = recurse

    def natural(self):
        return self.generate('natural')

    def ordered(self):
        v = self.generate('ordered')
        return self.apply_ordering(v)

    def tokenized(self, accum=None):
        if accum is None:
            accum = collections.OrderedDict()

        if self.recurse:
            v = self.apply_ordering(self.generate('tokenized', accum=accum))
        else:
            v = self.ordered()

        t = tokenize(v)

        accum[t] = v
        return tokenize(v)


class D(Node):
    def generate(self, op, **passthru):
        return dict((k, op_or_ident(v, op, **passthru)) for k, v in six.iteritems(self.vals))

    def apply_ordering(self, v):
        return collections.OrderedDict((k, v[k]) for k in sorted(v.keys()))


class L(Node):
    def generate(self, op, **passthru):
        return tuple(op_or_ident(v, op, **passthru) for v in self.vals)

    def apply_ordering(self, v):
        return list(v)

def p_ordered_dict(o, indent, outdent):
    return 'OrderedDict([\n' + \
                indent + \
                (',\n%s' % indent).join(pretty(kvp, indent) for kvp in six.iteritems(o)) + \
            '\n%s])' % outdent

def p_dict(o, indent, outdent):
    return '{\n' + \
                indent + \
                ('\n%s' % indent).join("%s: %s," % (pretty(k, indent), pretty(v, indent))
                                       for k, v in six.iteritems(o)) + \
            '\n%s}' % outdent

def p_list(o, indent, outdent):
    return '[\n' + \
                indent + \
                (',\n%s' % indent).join(pretty(v, indent) for v in o) + \
            '\n%s]' % outdent

def p_tuple(o, indent, outdent):
    if len(o) == 1:
        return '(%s,)' % pretty(o[0], indent)
    return '(\n' + \
                indent + \
                (',\n%s' % indent).join(pretty(v, indent) for v in o) + \
            '\n%s)' % outdent

def p_tokenset(o, indent, outdent):
    return 'Tokenset(\n%s%s,\n%stoken=%s\n%s)' % (
            indent,
            pretty(o.tokens, indent),
            indent,
            pretty(o.token, indent),
            outdent)


PRINTERS = (
        (Tokenset, p_tokenset),
        (OrderedDict, p_ordered_dict),
        (dict, p_dict),
        (list, p_list),
        (tuple, p_tuple),
    )


def pretty(struct, indent=''):
    outdent = indent
    indent = outdent + '    '
    for type_, handler in PRINTERS:
        if isinstance(struct, type_):
            return handler(struct, indent, outdent)
    return repr(struct)


def name(var, suffix):
    return '%s_%s' % (var.upper(), suffix.upper())


def constant(var, graph):
    return '%s = %s' % (var, pretty(graph))



def variable_set(varname, graph):
    yield constant(name(varname, 'natural'), graph.natural())
    yield constant(name(varname, 'ordered'), graph.ordered())
    tokens = collections.OrderedDict()
    graph.tokenized(tokens)
    yield constant(name(varname, 'tokens'), Tokenset(tokens, list(tokens.keys())[-1]))


def variables():
    dict_a = D({"a": "aye!", "A": "Eh."})
    yield ('d_a', dict_a)

    yield ('d_b',
           D({"b": "bi", "B": "Bee!"}))

    yield ('l_d_a', L([dict_a]))

    d_a_in_b = D({"b": "b", "a": dict_a}, recurse=False)
    yield 'd_a_in_b', d_a_in_b

    yield 'l_d_a_in_b', L([d_a_in_b])

    yield ('l_nest_partial',
           L([
               D({"abc": 123, "bcd": 234}, recurse=False),
               D({"bro": L([D({"name": "johnny", "gender": "male"}, recurse=False)]),
                  "sis": L([D({"name": "sissy", "gender": "female"}, recurse=False)])}),
            ]))

    yield ('l_multi_nest_partial',
           L([
               D({123: L(["one", "two", "three"])}, recurse=False),
               D({"1-0": L([
                    D({"a dict": D({"a": "b"}), "a list": L(["a", "b"])}, recurse=False),
                    D({"2-0": L([D({"eins": 1, "zwei": 2}, recurse=False)]),
                        "2-1": L([D({"foo": "bar", "baz": L(["blah"])}, recurse=False)]),
                    }),
                  ]),
                  "1-1": L([
                    D({"cracktopolis": "methboro", "inner": D({"gar": "bage"})}, recurse=False),
                    D({"2-0": L([D({"some": "thing", "inner": D({"or": L(["other"])})}, recurse=False)]),
                        "2-1": L([D({"goat": "milk"}, recurse=False)]),
                    }),
                  ]),
                }),
           ]))

    yield 'readme_full_transform', D({"first": L(["a", "b", "c"]),
                                      "second": D({"first": "1st!", "second": "2nd!"})})

    yield 'tokendict_case', D({"a": D({
                                        "a-0": L(["a0", "a1", "a2"]),
                                        "a-1": D({"a1": "A1", "a2": "A2"}),
                                    }, recurse=False),
                                "b": L([
                                        D({"b-0-0": "b00", "b-0-1": "b01"}),
                                        L(["b-1-0", "b-1-1", "b-1-2"]),
                                    ], recurse=False),
                                "c": L(["c0", "c1", "c2"], recurse=False),
                                "d": D({"d-0": "D0", "d-1": "D1", "d-2": "D2"}, recurse=False),
                            })

    yield ('walker_list_case',
            L([
                "string",
                57,
                L(["string", 30, D({"a": "b"}), L(["a", "b", "c"])]),
                D({"string": "stringy",
                    "number": 20,
                    "dict": D({"a": "b"}),
                    "list": L(["another", "list"])})
            ]))

    yield ('walker_dict_case',
            D({
                "string": "stringy",
                "number": 30000,
                "dict": D({
                        "string": "stringy",
                        "number": -52,
                        "dict": D({"a": "b"}),
                        "list": L(["another", "list"]),
                    }),
                "list": L([
                        "string",
                        23423,
                        D({
                            "string": "stringy",
                            "number": -12,
                            "dict": D({"a": "b"}),
                            "list": L(["another", "list"]),
                            }),
                        L([
                            "string",
                            1234,
                            D({"a": "b"}),
                            L(["another", "list"]),
                            ]),
                    ]),
            }))

def generate(stream):
    six.print_(COMMON_CODE, file=stream)
    for name, structure in variables():
        for definition in variable_set(name, structure):
            six.print_(definition, file=stream)

if __name__ == '__main__':
    generate(sys.stdout)




from merky import util
from merky import transformer

__all__ = ('Transformer',
           'AnnotationTransformer',
           'ExcludeAnnotationTransformer',
           'annotate',
          )

annotate = util.annotate
Transformer = transformer.Transformer
AnnotationTransformer = transformer.AnnotationTransformer
ExcludeAnnotationTransformer = transformer.ExcludeAnnotationTransformer




import collections

from nose import tools
from . import words
import merky

# Need to sort out how to do this with six compatibility layer
BASIC_JSON = words.unify('{"false":false,"negative integer":-17,"positive integer":25000,"string":"some string","true":true,"unicode":"',
                         words.SHEKELS,
                         '"}')

# The result of hashlib.sha1(BASIC_JSON.encode("utf-8")).hexdigest()
BASIC_HASH = '1fdd21a94597f8df08e75f67100e1fdcf5714a14'

BASIC = {"string": "some string",
         "true": True,
         "false": False,
         "positive integer": 25000,
         "negative integer": -17,
         "unicode": words.SHEKELS,
        }

BASIC_ORDERED = collections.OrderedDict((
                    ('false', False),
                    ('negative integer', -17),
                    ('positive integer', 25000),
                    ('string', 'some string'),
                    ('true', True),
                    ('unicode', words.SHEKELS),
                ))

NESTING_DICT = {
        "another string": "more strings!",
        "honesty": True,
        "lies": False,
        "a": dict(BASIC),
        "c": dict(BASIC),
       }

NESTING_DICT_ORDERED = collections.OrderedDict((
                            ('a', BASIC_HASH),
                            ('another string', 'more strings!'),
                            ('c', BASIC_HASH),
                            ('honesty', True),
                            ('lies', False),
                       ))

NESTING_DICT_JSON = words.unify('{',
                                  '"a":"', BASIC_HASH, '",',
                                  '"another string":"more strings!"', ',',
                                  '"c":"', BASIC_HASH, '",'
                                  '"honesty":true,',
                                  '"lies":false',
                                '}')

# hashlib.sha1(NESTING_DICT_JSON.encode("utf-8")).hexdigest()
NESTING_DICT_HASH = '593a122791ea25479cc6b2c16d69a5fb81241edb'

LIST = ["blah", words.ANGSTROM, words.EUROS]
LIST_JSON = words.unify('["blah",',
                        '"', words.ANGSTROM, '",',
                        '"', words.EUROS, '"',
                        ']')
LIST_HASH = '447f4779bf4a9865c600da63256eab9c827e48b0'

NESTING_LIST = ["something",
                dict(BASIC),
                list(LIST),
                words.EPEES]

NESTING_LIST_JSON = words.unify('[',
                                '"something",'
                                '"', BASIC_HASH, '",',
                                '"', LIST_HASH, '",',
                                '"', words.EPEES, '"',
                                ']')

NESTING_LIST_HASH = 'd724e888872c0a0545b13c0f16980953df360bf9'

def test_string():
    t = merky.Transformer()
    r = t.transform("what?")
    tools.assert_equal(("74417e40273f06624262871962bd9e738162da47", "what?"), next(r))
    tools.assert_raises(StopIteration, next, r)

def test_integer():
    t = merky.Transformer()
    r = t.transform(54321)
    tools.assert_equal(("348162101fc6f7e624681b7400b085eeac6df7bd", 54321), next(r))
    tools.assert_raises(StopIteration, next, r)

def test_basic_dict():
    t = merky.Transformer()
    r = t.transform(BASIC)
    tools.assert_equal((BASIC_HASH, BASIC_ORDERED), next(r))
    tools.assert_raises(StopIteration, next, r)

def test_nesting_dict():
    t = merky.Transformer()
    r = t.transform(NESTING_DICT)
    # We know the order because its depth-first traversal.
    # We also expect the hash to be the same for each copy of BASIC.

    # For the "a" member.
    tools.assert_equal((BASIC_HASH, BASIC_ORDERED), next(r))
    # For the "c" member.
    tools.assert_equal((BASIC_HASH, BASIC_ORDERED), next(r))
    # For the top container.  Note substituion of hashes.
    tools.assert_equal((NESTING_DICT_HASH, dict(NESTING_DICT, a=BASIC_HASH, c=BASIC_HASH)), next(r))
    tools.assert_raises(StopIteration, next, r)

def test_basic_list():
    t = merky.Transformer()
    r = t.transform(LIST)
    tools.assert_equal((LIST_HASH, LIST), next(r))
    tools.assert_raises(StopIteration, next, r)

def test_nesting_list():
    t = merky.Transformer()
    r = t.transform(NESTING_LIST)

    tools.assert_equal((BASIC_HASH, BASIC_ORDERED), next(r))
    tools.assert_equal((LIST_HASH, LIST), next(r))
    tools.assert_equal((NESTING_LIST_HASH, ["something", BASIC_HASH, LIST_HASH, words.EPEES]), next(r))
    tools.assert_raises(StopIteration, next, r)

